@@ -3,19 +3,7 @@
 #ifndef _BASHJMP_H_
 #define _BASHJMP_H_
 
-#include <setjmp.h>
-
-/* This *must* be included *after* config.h */
-
-#if defined (HAVE_POSIX_SIGSETJMP)
-#  define procenv_t	sigjmp_buf
-#  undef setjmp
-#  define setjmp(x)	sigsetjmp((x), 1)
-#  undef longjmp
-#  define longjmp(x, n)	siglongjmp((x), (n))
-#else
-#  define procenv_t	jmp_buf
-#endif
+#include "posixjmp.h"
 
 extern procenv_t	top_level;
 extern procenv_t	subshell_top_level;
